from logging import getLogger

from sqlalchemy.orm import aliased
from sqlalchemy.sql.functions import func

from ..names import ALL
from ..lib import time_to_local_time
from ..lib.log import Record
from ..sql import ActivityTopicJournal, FileHash, FileScan, StatisticJournal, ActivityTopic, ActivityTopicField, \
    StatisticName, ActivityJournal, ActivityGroup

log = getLogger(__name__)


def check(args, system, db):
    '''
## check

    > ch2 check

This is still in development.
    '''
    record = Record(log)
    check_all(record, db)


def check_all(record, db):
    check_activity_diary(record, db)


def check_activity_diary(record, db):
    check_activity_diary_missing_files(record, db)
    check_inconsistent_groups_activity_journal_v_topic(record, db)
    check_inconsistent_groups_activity_journal_v_topic_statistics(record, db)
    check_activity_topic_multiple_groups(record, db)


def check_activity_diary_missing_files(record, db):
    with record.record_exceptions(catch=True):
        with db.session_context() as s:
            bad = False
            for topic_journal in s.query(ActivityTopicJournal). \
                    join(StatisticJournal, StatisticJournal.source_id == ActivityTopicJournal.id). \
                    join(FileHash). \
                    outerjoin(FileScan). \
                    filter(FileScan.path == None).all():
                bad = True
                record.warning(f'{ActivityTopicJournal.__table__} with file hash '
                               f'{topic_journal.file_hash.hash[:6]} has associated entries but no activity')
                for statistic in s.query(StatisticJournal). \
                        filter(StatisticJournal.source == topic_journal).all():
                    record.info(f"{statistic.statistic_name.name}: {statistic.value}")
            if bad:
                record.info(f'Manual fix: try to infer missing activity from entries and upload file')
            else:
                record.info('No missing activities from activity topic data')


def check_inconsistent_groups_activity_journal_v_topic(record, db):
    with record.record_exceptions(catch=True):
        with db.session_context() as s:
            bad = False
            group_via_activity = aliased(ActivityGroup)
            group_via_topic = aliased(ActivityGroup)
            # this is tying journal entries to the activities and the topics.  since these are only
            # connected via file hash (since they need to persist across database reloads) they can
            # be inconsistent.
            for tjournal, agroup, tgroup, fscan in s.query(
                    ActivityTopicJournal, group_via_activity, group_via_topic, FileScan). \
                    join(StatisticJournal, StatisticJournal.source_id == ActivityTopicJournal.id). \
                    join(FileHash). \
                    join(FileScan). \
                    join(ActivityJournal, ActivityJournal.file_hash_id == FileHash.id). \
                    join(group_via_activity, ActivityJournal.activity_group_id == group_via_activity.id). \
                    join(StatisticName). \
                    join(ActivityTopicField). \
                    join(ActivityTopic). \
                    join(group_via_topic, ActivityTopic.activity_group_id == group_via_topic.id). \
                    filter(group_via_topic != group_via_activity).all():
                bad = True
                record.warning(f'{fscan.path} has {ActivityTopic.__table__} for {group_via_topic.name} but '
                               f'{ActivityJournal.__table__} for {group_via_activity.name}')
            if bad:
                record.info('Manual fix: rebuild with correct kit (if kit is determining activity group)')
            else:
                record.info('No inconsistent groups for activity topics')


def check_inconsistent_groups_activity_journal_v_topic_statistics(record, db):
    with record.record_exceptions(catch=True):
        with db.session_context() as s:
            q = s.query(StatisticName, ActivityJournal, StatisticJournal). \
                join(StatisticJournal, StatisticJournal.statistic_name_id == StatisticName.id). \
                join(ActivityTopicJournal, StatisticJournal.source_id == ActivityTopicJournal.id). \
                join(ActivityJournal, ActivityTopicJournal.file_hash_id == ActivityJournal.file_hash_id). \
                filter(StatisticName.activity_group_id != ActivityJournal.activity_group_id). \
                order_by(ActivityJournal.start)
            # log.debug(q)
            for name, activity, journal in q.all():
                q = s.query(StatisticJournal). \
                    join(StatisticName, StatisticName.id == StatisticJournal.statistic_name_id). \
                    join(ActivityTopicJournal, StatisticJournal.source_id == ActivityTopicJournal.id). \
                    join(ActivityJournal, ActivityJournal.file_hash_id == ActivityTopicJournal.file_hash_id). \
                    filter(StatisticName.name == name.name,
                           StatisticName.activity_group_id == ActivityJournal.activity_group_id,
                           ActivityJournal.id == activity.id)
                # log.debug(q)
                alternate = q.one_or_none()
                if alternate:
                    record.warning(f'Activity on {time_to_local_time(activity.start)} '
                                   f'for group {activity.activity_group.name} '
                                   f'is associated with journal statistic {name.name} '
                                   f'for group {name.activity_group.name} '
                                   f'with value {journal.value} and correct alternative {alternate.value}')
                else:
                    record.warning(f'Activity on {time_to_local_time(activity.start)} '
                                   f'for group {activity.activity_group.name} '
                                   f'is associated with journal statistic {name.name} '
                                   f'for group {name.activity_group.name} '
                                   f'with value {journal.value} and no alternative')


def check_activity_topic_multiple_groups(record, db):
    with record.record_exceptions(catch=True):
        with db.session_context() as s:
            q = s.query(func.group_concat(ActivityGroup.name.distinct()), ActivityTopicJournal). \
                join(StatisticJournal, StatisticJournal.source_id == ActivityTopicJournal.id). \
                join(StatisticName, StatisticName.id == StatisticJournal.statistic_name_id). \
                filter(ActivityGroup.id == StatisticName.activity_group_id,
                       ActivityTopicJournal.file_hash_id == ActivityJournal.file_hash_id,
                       ActivityGroup.name != ALL). \
                group_by(ActivityTopicJournal)
            log.debug(q)
            # tried to query journal as well as topic and hit sqlalchemy bug, so order in python
            all = sorted(q.all(), key=lambda x: x[1].file_hash.activity_journal.start)
            for (groups, topic) in all:
                journal = topic.file_hash.activity_journal
                if ',' in groups:
                    record.warning(f'Activity topic journal for {time_to_local_time(journal.start)} '
                                   f'is associated with multiple groups: {groups}')

# add flag fix to fix things

# check everything needed for power estimates
# (including weight)

# check constants defined

# check configure was run (version)

# check whether permanent directory present

# check whether file system ok (space?)

# check whether all files are scanned

# check whether all activity topics entries have an activity for the file hash

# all activity diary entries have corresponding activity - fix group

# check for activity without GPS datya (lyn in 2018)

# check for multiple activiy topic journals (or whatever we have)



from logging import getLogger

from sqlalchemy import func

from ..json import JsonResponse
from ...commands.search import expand_activities, unified_search
from ...lib.utils import parse_bool
from ...sql import StatisticName, ActivityGroup, StatisticJournal, Source
from ...sql.tables.source import SourceType

log = getLogger(__name__)

NAME = 'name'
DESCRIPTION = 'description'
GROUPS = 'groups'
UNITS = 'units'

RESULTS = 'results'
ERROR = 'error'


class Search:

    @staticmethod
    def query_activity(request, s, query):
        try:
            advanced = parse_bool(request.args.get('advanced', 'false'), default=None)
            log.info(f'{"advanced " if advanced else ""}query: {query}')
            return JsonResponse({RESULTS: expand_activities(s, unified_search(s, query, advanced=advanced))})
        except Exception as e:
            log.warning(e)
            return JsonResponse({ERROR: str(e)})

    @staticmethod
    def read_activity_terms(request, s):

        def format(row):
            name, description,units,  groups = row
            groups = ', '.join(sorted(groups.split(',')))
            return {NAME: name, DESCRIPTION: description, GROUPS: groups, UNITS: units}

        q = s.query(StatisticName.name, StatisticName.description, StatisticName.units,
                    func.group_concat(ActivityGroup.name.distinct())). \
            join(ActivityGroup). \
            join(StatisticJournal). \
            join(Source, Source.id == StatisticJournal.source_id). \
            filter(Source.type.in_([SourceType.ACTIVITY_TOPIC, SourceType.ACTIVITY])). \
            group_by(StatisticName.name). \
            order_by(StatisticName.name)
        log.debug(q)
        return JsonResponse([format(row) for row in q.all()])




import datetime as dt
from logging import getLogger

from sqlalchemy import Column, Text, Integer, ForeignKey, UniqueConstraint, desc
from sqlalchemy.orm import relationship, backref

from .source import Source, SourceType
from ..support import Base
from ..types import Time, Sort, ShortCls, NullStr
from ...lib.date import format_time, local_date_to_time, local_time_to_time

log = getLogger(__name__)


class ActivityGroup(Base):

    __tablename__ = 'activity_group'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False, server_default='')
    description = Column(Text, nullable=False, server_default='')
    sort = Column(Sort, nullable=False, server_default='')

    def __str__(self):
        return self.name

    @classmethod
    def from_name(cls, s, name):
        if not name:
            raise Exception('Missing activity group (None)')
        if isinstance(name, ActivityGroup):
            return name  # allow callers to already have an instance
        else:
            return s.query(ActivityGroup).filter(ActivityGroup.name.ilike(name)).one()


class ActivityJournal(Source):

    __tablename__ = 'activity_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    activity_group_id = Column(Integer, ForeignKey('activity_group.id', ondelete='cascade'), nullable=False)
    activity_group = relationship('ActivityGroup')
    file_hash_id = Column(Integer, ForeignKey('file_hash.id'), nullable=False)
    file_hash = relationship('FileHash', backref=backref('activity_journal', uselist=False))
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    UniqueConstraint(activity_group_id, start)
    UniqueConstraint(file_hash_id)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.ACTIVITY
    }

    def __str__(self):
        return 'ActivityJournal %s %s to %s' % (self.activity_group.name,
                                                format_time(self.start), format_time(self.finish))

    def time_range(self, s):
        return self.start, self.finish

    def get_named(self, s, qname, owner=None):
        from .. import StatisticJournal, StatisticName
        from ...data.constraint import parse_qname
        name, group = parse_qname(qname)
        q = s.query(StatisticJournal). \
            join(ActivityJournal). \
            join(StatisticName). \
            filter(StatisticName.name.ilike(name),
                   StatisticJournal.source_id == self.id)
        if owner: q = q.filter(StatisticName.owner == owner)
        if group: q = q.join(ActivityGroup).filter(ActivityGroup.name.ilike(group))
        elif group is None: q = q.join(ActivityGroup).filter(ActivityGroup.id == ActivityJournal.activity_group_id)
        return q.all()

    def get_all_named(self, s, qname, owner=None):
        return self.get_named(s, qname, owner=owner) + \
               self.get_activity_topic_journal(s).get_named(s, qname, owner=owner)

    def get_activity_topic_journal(self, s):
        from .. import ActivityTopicJournal, FileHash
        return s.query(ActivityTopicJournal). \
            join(FileHash). \
            join(ActivityJournal). \
            filter(ActivityJournal.id == self.id).one()

    @classmethod
    def at_date(cls, s, date):
        day = local_date_to_time(date)
        return s.query(ActivityJournal).filter(ActivityJournal.start >= day,
                                               ActivityJournal.start < day + dt.timedelta(days=1)).all()

    @classmethod
    def at_local_time(cls, s, local_time):
        time = local_time_to_time(local_time)
        return s.query(ActivityJournal).filter(ActivityJournal.start == time).one()

    @classmethod
    def before_local_time(cls, s, local_time):
        time = local_time_to_time(local_time)
        return s.query(ActivityJournal). \
            filter(ActivityJournal.start < time). \
            order_by(desc(ActivityJournal.start)). \
            limit(1).one_or_none()

    @classmethod
    def after_local_time(cls, s, local_time):
        time = local_time_to_time(local_time)
        return s.query(ActivityJournal). \
            filter(ActivityJournal.start >= (time + dt.timedelta(days=1))). \
            order_by(ActivityJournal.start). \
            limit(1).one_or_none()

    @classmethod
    def number_of_activities(cls, s):
        return s.query(ActivityJournal).count()


class ActivityTimespan(Base):

    __tablename__ = 'activity_timespan'

    id = Column(Integer, primary_key=True)
    activity_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                                 nullable=False, index=True)
    activity_journal = relationship('ActivityJournal',
                                    backref=backref('timespans', cascade='all, delete-orphan',
                                                    passive_deletes=True,
                                                    order_by='ActivityTimespan.start'))
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    UniqueConstraint(activity_journal_id, start)

    def __str__(self):
        return 'ActivityTimespan from %s - %s' % (format_time(self.start), format_time(self.finish))


class ActivityBookmark(Base):

    __tablename__ = 'activity_bookmark'

    id = Column(Integer, primary_key=True)
    activity_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), nullable=False)
    activity_journal = relationship('ActivityJournal')
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    owner = Column(ShortCls, nullable=False, index=True)  # index for deletion
    constraint = Column(NullStr, index=True)
    UniqueConstraint(activity_journal_id, start, finish, owner, constraint)

    def __str__(self):
        return 'ActivityBookmark from %s - %s' % (format_time(self.start), format_time(self.finish))





from .nearby import fmt_nearby, nearby_any_time
from ..display import ActivityJournalDelegate
from ....diary.model import text, link, optional_text, COMPARE_LINKS
from ....lib import time_to_local_time
from ....lib.date import format_date


class JupyterDelegate(ActivityJournalDelegate):

    @optional_text('Jupyter', tag='jupyter-activity')
    def read_journal_date(self, s, ajournal, date):
        links = [link('None', db=(time_to_local_time(ajournal.start), None, ajournal.activity_group.name))] + \
                [link(fmt_nearby(ajournal2, nb),
                      db=(time_to_local_time(ajournal.start), time_to_local_time(ajournal2.start), ajournal.activity_group.name))
                 for ajournal2, nb in nearby_any_time(s, ajournal)]
        yield [text('Compare to', tag=COMPARE_LINKS)] + links
        yield link('All Similar', db=(time_to_local_time(ajournal.start), ajournal.activity_group.name))

    @optional_text('Jupyter', tag='jupyter-activity')
    def read_schedule(self, s, date, schedule):
        finish = schedule.next_frame(date)
        yield link('All Activities', db=(format_date(date), format_date(finish)))




import datetime as dt
from logging import getLogger

from urwid import MainLoop, connect_signal

from .args import DATE, SCHEDULE, FAST, mm, BASE
from ..diary.database import read_date, read_schedule
from ..diary.model import COMPARE_LINKS
from ..diary.views.urwid import build, layout_date, layout_schedule
from ..jupyter.server import set_controller, JupyterController
from ..jupyter.template.activity_details import activity_details
from ..jupyter.template.all_activities import all_activities
from ..jupyter.template.compare_activities import compare_activities
from ..jupyter.template.health import health
from ..jupyter.template.similar_activities import similar_activities
from ..lib.date import to_date, local_time_to_time
from ..lib.io import tui
from ..lib.schedule import Schedule
from ..lib.utils import PALETTE
from ..lib.widgets import DateSwitcher
from ..pipeline.pipeline import run_pipeline
from ..pipeline.display.activity.nearby import NEARBY_LINKS
from ..sql import PipelineType, DiaryTopicJournal, ActivityJournal
from ..sql.database import StatisticJournal
from ..urwid.tui.factory import Factory
from ..urwid.tui.tabs import TabList

log = getLogger(__name__)


@tui
def diary(args, system, db):
    '''
## diary

    > ch2 diary [DATE]

The date can be an absolute day or the number of days previous.  So `ch2 diary 1` selects yesterday.

Display the daily diary.  Enter information here.

To exit, alt-q (or, without saving, alt-x).

    > ch2 diary (--month | --year | --schedule SCHEDULE) [DATE}

Display a summary for the month / year / schedule.
    '''
    set_controller(JupyterController(args, system))
    date, schedule, base = args[DATE], args[SCHEDULE], args[BASE]
    if not date:
        date = dt.date.today()
    else:
        days = None
        try:
            # try int first because we need to separate the case of days from years
            days = int(date)
            if days > 1000:
                days = None
        except ValueError:
            pass
        if days is None:
            date = to_date(date)
        else:
            date = dt.date.today() - dt.timedelta(days=days)
    with db.session_context() as s:
        DiaryTopicJournal.check_tz(system, s)
    if schedule:
        schedule = Schedule(schedule)
        if schedule.start or schedule.finish:
            raise Exception('Schedule must be open (no start or finish)')
        MainLoop(ScheduleDiary(db, date, schedule), palette=PALETTE).run()
    else:
        MainLoop(DailyDiary(db, date), palette=PALETTE).run()
        if not args[FAST]:
            print('\n  Please wait while statistics are updated...')
            run_pipeline(system, db, base, PipelineType.CALCULATE)
            print(f'  ...done (thanks! - use {mm(FAST)} to avoid this, if the risk is worth it)\n')


class Diary(DateSwitcher):

    def __init__(self, db, date):
        super().__init__(db, date)

    def save(self):
        s = self._session
        if s:
            self.__clean(s, s.dirty, delete=True)
            self.__clean(s, s.new, delete=False)
        super().save()

    @staticmethod
    def __clean(s, instances, delete=False):
        for instance in instances:
            if isinstance(instance, StatisticJournal):
                if instance.value == None:
                    log.debug(f'Discarding {instance}')
                    if delete:
                        s.delete(instance)
                    else:
                        s.expunge(instance)

    def _wire(self, active, name, callback):
        if name in active:
            for menu in active[name]:
                connect_signal(menu, 'click', callback)
            del active[name]


class DailyDiary(Diary):
    '''
    Render the diary at a given date.
    '''

    def _build(self, s):
        log.debug('Building diary at %s' % self._date)
        model = list(read_date(s, self._date))
        f = Factory(TabList())
        active, widget = build(model, f, layout_date)
        self._wire(active, NEARBY_LINKS, lambda m: self._change_date(to_date(m.state[0].split(' ')[0])))
        self._wire(active, COMPARE_LINKS, lambda m: self.__show_gui(*m.state))
        self._wire(active, 'health', lambda l: self.__show_health())
        self._wire(active, 'all-similar', lambda l: self.__show_similar(l.state[0]))
        if active:
            raise Exception(f'Unhandled links: {", ".join(active.keys())}')
        return widget, f.tabs

    def __show_gui(self, aj1, aj2, group):
        aj1t, aj2t = local_time_to_time(aj1), local_time_to_time(aj2) if aj2 else None
        aj1 = ActivityJournal.at_local_time(self._session, aj1)
        if aj2:
            compare_activities(aj1t, aj2t, aj1.activity_group.name)
        else:
            activity_details(aj1t, aj1.activity_group.name)

    def __show_similar(self, local_time):
        aj1 = ActivityJournal.at_local_time(self._session, local_time)
        similar_activities(aj1.start, aj1.activity_group.name)

    def __show_health(self):
        health()


class ScheduleDiary(Diary):
    '''
    Display summary data for the given schedule.
    '''

    def __init__(self, db, date, schedule):
        self._schedule = schedule
        super().__init__(db, schedule.start_of_frame(date))

    def _build(self, s):
        log.debug(f'Building diary at {self._date} for {self._schedule}')
        model = list(read_schedule(s, self._schedule, self._date))
        f = Factory(TabList())
        active, widget = build(model, f, layout_schedule)
        self._wire(active, 'all-activities', lambda l: self.__show_all())
        if active:
            raise Exception(f'Unhandled links: {", ".join(active.keys())}')
        return widget, f.tabs

    def __show_all(self):
        finish = self._schedule.next_frame(self._date)
        all_activities(self._date.strftime('%Y-%m-%d'), finish.strftime('%Y-%m-%d'))



from logging import getLogger

from sqlalchemy import intersect, or_

from .args import QUERY, SHOW, SET, ADVANCED
from ..config.config import NOTES
from ..data import constrained_activities
from ..data.constraint import constraint, check_constraint
from ..diary.model import DB, VALUE, UNITS
from ..lib import time_to_local_time
from ..sql import ActivityTopicJournal, FileHash, ActivityJournal, StatisticJournal, ActivityTopicField, ActivityTopic, \
    StatisticJournalText, StatisticName
from ..pipeline.calculate.activity import ActivityCalculator
from ..names import TIME, START, ACTIVE_TIME, DISTANCE, ACTIVE_DISTANCE, GROUP

log = getLogger(__name__)

name = ActivityTopicField.NAME.lower()
group = GROUP.lower()
start = START.lower()
time = TIME.lower()
distance = DISTANCE.lower()


def search(args, system, db):
    '''
## search

    > ch2 search [-a|--advanced] QUERY [--show NAME ...] [--set NAME=VALUE]

This searches for activities.
Once a matching activity is found additional statistics can be displayed (--show)
and a single value modified (--set).

Simple searches (without --advanced) look for matches of all given words in the
name and notes fields for the activity.

The advanced syntax is similar to SQL, but element names are statistic names.
The name can include the activity group (start:bike) and SQL wildcards (%fitness%).
A name of the form "name:" matches any activity group;
"name" matches the activity group of the matched activity
(usually what is needed - the main exception is some statistics defined for group All).

For advanced searches string values must be quoted, negation and NULL values are not supported,
and comparison must be between a name and a value (not two names).

### Example

    > ch2 search --advanced 'name="Wrong Name"' --set 'name="Right Name"'

    '''
    query, show, set, advanced = args[QUERY], args[SHOW], args[SET], args[ADVANCED]
    if not show and not set: show = [ACTIVE_TIME, ACTIVE_DISTANCE]
    with db.session_context() as s:
        activities = unified_search(s, query, advanced=advanced)
        process_activities(s, activities, show, set)


def unified_search(s, query, advanced=True):
    if advanced:
        return constrained_activities(s, query)
    else:
        return simple_activity_search(s, query)


def expand_activities(s, activities):
    return [expand_activity(s, activity) for activity in activities]


def expand_activity(s, activity_journal):
    topic_journal = s.query(ActivityTopicJournal). \
        join(FileHash).join(ActivityJournal). \
        filter(ActivityJournal.id == activity_journal.id).one()

    def format(statistic):
        if statistic:
            return {VALUE: statistic.value, UNITS: statistic.statistic_name.units}
        else:
            return None

    def value(statistic):
        if statistic:
            return statistic.value
        else:
            return None

    return {DB: activity_journal.id,
            name: {VALUE: value(StatisticJournal.for_source(s, topic_journal.id, ActivityTopicField.NAME,
                                                            ActivityTopic, activity_journal.activity_group)),
                   UNITS: None},
            group: {VALUE: activity_journal.activity_group.name, UNITS: None},
            start: {VALUE: time_to_local_time(activity_journal.start), UNITS: 'date'},
            time: format(StatisticJournal.for_source(s, activity_journal.id, ACTIVE_TIME,
                                                     ActivityCalculator, activity_journal.activity_group)),
            distance: format(StatisticJournal.for_source(s, activity_journal.id, ACTIVE_DISTANCE,
                                                         ActivityCalculator, activity_journal.activity_group))}


def display(data):
    if data:
        value = data[0].formatted()
        if len(data) > 1: value += '...'
        return value
    else:
        return '-'


def parse_set(s, set):
    name, op, value = constraint(set)[0]
    if op != '=': raise Exception(f'{set} must be simple assignment')
    check_constraint(s, (name, op, value))
    return name, value


def process_activities(s, activities, show, set):
    if set:
        name, value = parse_set(s, set)
    for activity in activities:
        print(f'{display(activity.get_named(s, START, owner=ActivityCalculator))}  {activity.activity_group.name}  '
              f'{display(activity.get_all_named(s, ActivityTopicField.NAME, owner=ActivityTopic))}')
        for qname in show:
            for result in activity.get_all_named(s, qname):
                print(f'  {result.statistic_name.name} = {result.formatted()}')
        if set:
            for result in activity.get_all_named(s, name):
                before = result.formatted()
                result.set(value)
                after = result.formatted()
                log.info(f'  {result.statistic_name}: {before!r} -> {after!r}')


def parse(query):
    word, quote = '', None
    for c in query:
        if c == ' ':
            if quote:
                word += c
            else:
                if word:
                    yield word
                    word = ''
        elif c == quote:
            quote = None
            yield word
            word = ''
        elif not word and c in ('"', "'"):
            quote = c
        else:
            word += c
    if quote:
        raise Exception('Unbalanced quote')
    if word:
        yield word


def word_query(s, ids, word):
    return s.query(StatisticJournalText.source_id). \
        filter(StatisticJournalText.statistic_name_id.in_(ids),
               StatisticJournalText.value.ilike('%' + word + '%'))


def simple_activity_search(s, query):
    words = list(parse(query))
    log.debug(f'Parsed {query} as {words}')
    words_query = None
    ids = [id[0] for id in s.query(StatisticName.id).
        filter(StatisticName.name.in_([NOTES, ActivityTopicField.NAME]),
               StatisticName.owner == ActivityTopic).all()]
    for word in words:
        if words_query is not None:
            words_query = intersect(words_query, word_query(s, ids, word)).select()
        else:
            words_query = word_query(s, ids, word)
    q = s.query(ActivityJournal). \
        join(FileHash). \
        join(ActivityTopicJournal). \
        filter(or_(ActivityJournal.id.in_(words_query),
                   ActivityTopicJournal.id.in_(words_query)))
    log.debug(q)
    return q.all()




import datetime as dt
from logging import getLogger
from re import escape

from sqlalchemy import union, intersect, not_

from ..lib import local_time_to_time
from ..lib.peg import transform, choice, pattern, sequence, Recursive, drop, exhaustive, single
from ..lib.utils import timing
from ..sql import ActivityJournal, StatisticName, StatisticJournalType, ActivityGroup, StatisticJournal, \
    ActivityTopicJournal, Source
from ..sql.tables.source import SourceType
from ..sql.tables.statistic import STATISTIC_JOURNAL_CLASSES

log = getLogger(__name__)


def pat(regexp, trans=None):
    parser = pattern(r'\s*' + regexp + r'\s*')
    if trans:
        parser = transform(parser, trans)
    return parser


def lit(text, trans=None):
    return pat('(' + escape(text) + ')', trans=trans)


def unexp(values):
    try:
        val, exp = values
        return [float(val) * 10 ** int(exp)]
    except:
        return [float(values[0])]


def to_tuple(l):
    return [tuple(l)]


def join(a, sep, b):
    return transform(sequence(a, sep, b), to_tuple)


def flip(l):
    value, op, name = l[0]
    return [(name, {'=': '=', '!=': '!=', '<': '>', '>': '<', '>=': '<=', '<=': '>='}[op], value)]


AND, OR = 'and', 'or'
name = pat(r'(\w(?:\s*\w)*(?::(?:\w(?:\s*\w)*)?)?)')
# important that the different value types are exclusive to avoid ambiguities
number = choice(pat(r'(-?\d+)', lambda x: [int(i) for i in x]),
                pat(r'(-?\d*\.\d+)', lambda x: [float(i) for i in x]),
                pat(r'(-?\d+\.\d*)', lambda x: [float(i) for i in x]),
                pat(r'([-+]?\d*\.?\d+)[eE]([-+]?\d+)', unexp))
string = choice(pat(r'"((?:[^"\\]|\\.)*)"'), pat(r"'((?:[^'\\]|\\.)*)'"))
date = r'\d{4}-\d{2}-\d{2}'
time = r'\d{2}(?::\d{2}(?::\d{2})?)?'
datetime = pat('(' + date + '(?:[T ]' + time + ')?)', lambda x: [local_time_to_time(t) for t in x])
value = choice(number, string, datetime)

operator = choice(*[lit(cmp) for cmp in ['=', '!=', '<', '>', '<=', '>=']])
comparison = choice(join(name, operator, value), transform(join(value, operator, name), flip))

term = Recursive()

and_comparison = Recursive()
and_comparison.calls(choice(term, join(term, lit(AND), and_comparison)))

or_comparison = Recursive()
or_comparison.calls(choice(and_comparison, join(and_comparison, lit(OR), or_comparison)))

parens = transform(sequence(drop(lit('(')), or_comparison, drop(lit(')'))))
term.calls(choice(comparison, parens))

constraint = single(exhaustive(choice(or_comparison, parens)))


def constrained_activities(s, query):
    with timing('parse AST'):
        ast = constraint(query)[0]
    log.debug(f'AST: {ast}')
    with timing('check AST'):
        check_constraints(s, ast)
    log.debug('Checked constraints')
    with timing('build SQL'):
        q = build_activity_query(s, ast)
    log.debug(f'Query: {q}')
    with timing('execute SQL'):
        return q.all()


def check_constraints(s, ast):
    l, op, r = ast
    if op in (AND, OR):
        check_constraints(s, l)
        check_constraints(s, r)
    else:
        check_constraint(s, ast)


def infer_types(value):
    if isinstance(value, str): return [StatisticJournalType.TEXT]
    if isinstance(value, dt.datetime): return [StatisticJournalType.TIMESTAMP]
    return [StatisticJournalType.FLOAT, StatisticJournalType.INTEGER]


def check_constraint(s, ast):
    qname, op, value = ast
    name, group = parse_qname(qname)
    q = s.query(StatisticName.id). \
        filter(StatisticName.name.ilike(name),
               StatisticName.statistic_journal_type.in_(infer_types(value)))
    if group:
        q = q.join(ActivityGroup).filter(ActivityGroup.name.ilike(group))
    if not q.count():
        raise Exception(f'No match for statistic {qname} with type {value.__class__.__name__}')
    if not s.query(StatisticJournal). \
            join(Source). \
            filter(StatisticJournal.statistic_name_id.in_(q),
                   Source.type.in_([SourceType.ACTIVITY, SourceType.ACTIVITY_TOPIC])).count():
        raise Exception(f'Statistic {qname} exists but is not associated with any activity data')


def build_activity_query(s, ast):
    constraints = build_constraints(s, ast).cte()
    return s.query(ActivityJournal). \
        filter(ActivityJournal.id.in_(constraints)). \
        order_by(ActivityJournal.start)


def build_constraints(s, ast):
    l, op, r = ast
    if op in (AND, OR):
        lcte = build_constraints(s, l)
        rcte = build_constraints(s, r)
        return build_join(op, lcte, rcte)
    else:
        return build_comparisons(s, ast)


def build_join(op, lcte, rcte):
    if op == OR:
        return union(lcte, rcte).select()
    else:
        return intersect(lcte, rcte).select()


def build_comparisons(s, ast):
    qname, op, value = ast
    name, group = parse_qname(qname)
    if isinstance(value, str):
        return get_activity_journal_ids(s, name, op, value, group,
                                        StatisticJournalType.TEXT, {'=': 'ilike', '!=': 'nlike'})
    elif isinstance(value, dt.datetime):
        return get_activity_journal_ids(s, name, op, value, group, StatisticJournalType.TIMESTAMP)
    else:
        qint = get_activity_journal_ids(s, name, op, value, group, StatisticJournalType.INTEGER)
        qfloat = get_activity_journal_ids(s, name, op, value, group, StatisticJournalType.FLOAT)
        return union(qint, qfloat).select()


def get_activity_journal_ids(s, name, op, value, group, type, update_attrs=None):
    statistic_journals = \
        get_statistic_journals(s, name, op, value, group, type, update_attrs=update_attrs).cte()
    via_activity_journal = s.query(ActivityJournal.id). \
        join(statistic_journals, ActivityJournal.id == statistic_journals.c.source_id)
    if group is None:
        via_activity_journal = \
            via_activity_journal.filter(ActivityJournal.activity_group_id == statistic_journals.c.activity_group_id)
    via_topic_journal = s.query(ActivityJournal.id). \
        join(ActivityTopicJournal, ActivityTopicJournal.file_hash_id == ActivityJournal.file_hash_id). \
        join(statistic_journals, ActivityTopicJournal.id == statistic_journals.c.source_id)
    if group is None:
        via_topic_journal = \
            via_topic_journal.filter(ActivityJournal.activity_group_id == statistic_journals.c.activity_group_id)
    return union(via_activity_journal, via_topic_journal).select()


def get_statistic_journals(s, name, op, value, group, type, update_attrs=None):
    attrs = {'=': '__eq__', '!=': '__ne__', '>': '__gt__', '>=': '__ge__', '<': '__lt__', '<=': '__le__'}
    if update_attrs:
        attrs.update(update_attrs)
    attr = attrs[op]
    statistic_journal = STATISTIC_JOURNAL_CLASSES[type]
    q = s.query(statistic_journal.source_id, StatisticName.activity_group_id). \
        join(StatisticName). \
        filter(StatisticName.name.ilike(name))
    if attr == 'nlike':  # no way to negate like in a single attribute
        q = q.filter(not_(statistic_journal.value.ilike(value)))
    else:
        q = q.filter(getattr(statistic_journal.value, attr)(value))
    if group:
        q = q.join(ActivityGroup).filter(ActivityGroup.name.ilike(group))
    return q


def parse_qname(qname):
    '''
    A qualified name (for a statistic) has the form NAME:GROUP with optional spaces where:

    NAME (with no qualifier) means 'match the group of the current activity' - returned as None
    NAME: (colon but no value) means 'match any constraint' - returned as ''
    NAME:GROUP means 'match the given group' - returned as the name
    '''
    if ':' in qname:
        name, group = qname.split(':')
        name, group = name.strip(), group.strip()
    else:
        name, group = qname.strip(), None
    log.debug(f'Parsed {qname} as {name}:{group}')
    return name, group




import datetime as dt
from logging import getLogger

import numpy as np
import pandas as pd
from sqlalchemy import desc, and_, or_, distinct, func, select
from sqlalchemy.sql.functions import count

from .read import AbortImport, AbortImportButMarkScanned, MultiProcFitReader
from .loader import StatisticJournalLoader
from ... import FatalException
from ...commands.args import MONITOR, mm, FORCE, DEFAULT
from ...data.frame import _tables
from ...fit.format.records import fix_degrees, unpack_single_bytes, merge_duplicates
from ...fit.profile.profile import read_fit
from ...lib.date import time_to_local_date, format_time
from ...names import HEART_RATE, BPM, STEPS, STEPS_UNITS, CUMULATIVE_STEPS, _new, TIME, SOURCE, ALL
from ...sql.database import StatisticJournalType, ActivityGroup
from ...sql.tables.monitor import MonitorJournal
from ...sql.tables.statistic import StatisticJournalInteger, StatisticName, StatisticJournal
from ...sql.utils import add

log = getLogger(__name__)

ACTIVITY_TYPE_ATTR = 'activity_type'
HEART_RATE_ATTR = 'heart_rate'
MONITORING_ATTR = 'monitoring'
MONITORING_INFO_ATTR = 'monitoring_info'
STEPS_ATTR = 'steps'


def missing_dates(s, force=False):
    # we don't try to be perfect here.  the idea is that it's called to get the latest
    # updates, rather than fill in all gaps (do the bulk download thing for that).
    # we also don't try to get fractional data.
    # and as for timezones... we just assume garmin uses the local timezone.
    latest = s.query(MonitorJournal).order_by(desc(MonitorJournal.start)).limit(1).one_or_none()
    if latest is None:
        log.warning('No existing monitor data - ' +
                    'do a bulk download instead: https://www.garmin.com/en-US/account/datamanagement/')
        return
    # find the mid-point to avoid any problems with timezones and edge cases
    # (these files don't span more than a day)
    seconds = (latest.finish - latest.start).total_seconds() / 2
    start = time_to_local_date(latest.start + dt.timedelta(seconds=seconds))
    finish = dt.date.today()
    days = (finish - start).days
    if days > 14 and not force:
        raise Exception('Too many days (%d) - ' % days +
                        'do a bulk download instead: https://www.garmin.com/en-US/account/datamanagement/')
    if days:
        while start < finish:
            yield start
            start += dt.timedelta(days=1)
    else:
        log.warning('No dates to download')


class MonitorLoader(StatisticJournalLoader):

    def _preload(self):
        dummy = super()._preload()
        try:
            for name in self._s.query(StatisticName). \
                    filter(StatisticName.name == CUMULATIVE_STEPS,
                           StatisticName.owner == self._owner).all():
                n = self._s.query(count(StatisticJournal.id)). \
                    filter(StatisticJournal.statistic_name == name,
                           StatisticJournal.time >= self.start,
                           StatisticJournal.time <= self.finish).scalar()
                if n and self.start and self.finish:
                    log.warning(f'Deleting {n} overlapping {CUMULATIVE_STEPS}')
                    self._s.query(StatisticJournal). \
                        filter(StatisticJournal.statistic_name == name,
                               StatisticJournal.time >= self.start,
                               StatisticJournal.time <= self.finish).delete()
        except:
            log.debug('Failed to clean database')
            self._s.rollback()
            raise
        return dummy

    def _resolve_duplicate(self, name, instance, prev):
        log.warning(f'Using max of duplicate values at {instance.time} for {name} ({instance.value}/{prev.value})')
        prev.value = max(prev.value, instance.value)


NEW_STEPS = _new(STEPS)
STEPS_DESCRIPTION = '''The increment in steps read from the FIT file.'''


class MonitorReader(MultiProcFitReader):

    def __init__(self, *args, sport_to_activity=None, **kargs):
        from ...commands.upload import MONITOR
        self.sport_to_activity = self._assert('sport_to_activity', sport_to_activity)
        super().__init__(*args, sub_dir=MONITOR, **kargs)

    def _startup(self, s):
        self.sport_to_activity_group = {label: group for label, group in self._expand_activity_groups(s)}
        super()._startup(s)

    def _expand_activity_groups(self, s):
        # extract default values from config that supports more complex activity loading
        for key, value in self.sport_to_activity.items():
            while value and not isinstance(value, str):
                try:
                    value = value[DEFAULT]
                except KeyError:
                    log.warning(f'Missing default in sport_to_activity for {key}')
                    value = None
            if value:
                yield key, ActivityGroup.from_name(s, value)

    def _get_loader(self, s, **kargs):
        if 'owner' not in kargs:
            kargs['owner'] = self.owner_out
        return MonitorLoader(s, **kargs)

    def _base_command(self):
        force = mm(FORCE) if self.force else ""
        return f'{MONITOR} {force}'

    def _delete_contained(self, s, start, finish, path):
        for mjournal in s.query(MonitorJournal). \
                filter(MonitorJournal.start >= start,
                       MonitorJournal.finish <= finish).all():
            log.warning(f'Replacing data from {mjournal.start} to {mjournal.finish} with data from '
                        f'{path} ({start} - {finish}')
            s.delete(mjournal)

    def _check_inside(self, s, start, finish, path):
        for mjournal in s.query(MonitorJournal). \
                filter(or_(and_(MonitorJournal.start <= start, MonitorJournal.finish > finish),
                           and_(MonitorJournal.start < start, MonitorJournal.finish >= finish))).all():
            log.warning('%s already includes data from %s for %s - %s' % (mjournal, path, start, finish))
            raise AbortImportButMarkScanned()

    def _check_overlap(self, s, start, finish, path):
        for mjournal in s.query(MonitorJournal). \
                filter(or_(and_(MonitorJournal.start < start,
                                MonitorJournal.finish > start, MonitorJournal.finish < finish),
                           and_(MonitorJournal.finish > finish,
                                MonitorJournal.start > start, MonitorJournal.start < finish))).all():
            log.warning('%s overlaps data from %s for %s - %s' % (mjournal, path, start, finish))
            raise AbortImport()

    def _delete_previous(self, s, start, finish, path):
        self._check_inside(s, start, finish, path)
        self._check_overlap(s, start, finish, path)
        self._delete_contained(s, start, finish, path)
        s.commit()

    @staticmethod
    def parse_records(data):
        return MonitorReader.read_fit_file(data, merge_duplicates, fix_degrees, unpack_single_bytes)

    @staticmethod
    def read_first_timestamp(path, records):
        return MonitorReader._first(path, records, MONITORING_INFO_ATTR).value.timestamp

    @staticmethod
    def read_last_timestamp(path, records):
        return MonitorReader._last(path, records, MONITORING_ATTR).value.timestamp

    def _read_data(self, s, file_scan):
        records = self.parse_records(read_fit(file_scan.path))
        first_timestamp = self.read_first_timestamp(file_scan.path, records)
        last_timestamp = self.read_last_timestamp(file_scan.path, records)
        if first_timestamp == last_timestamp:
            log.debug('File %s is empty (no timespan)' % file_scan)
            raise AbortImportButMarkScanned()
        if not first_timestamp:
            raise Exception('Missing timestamp in %s' % file_scan)

        log.info(f'Importing monitor data from {file_scan} '
                 f'for {format_time(first_timestamp)} - {format_time(last_timestamp)}')
        self._delete_previous(s, first_timestamp, last_timestamp, file_scan)
        mjournal = add(s, MonitorJournal(start=first_timestamp, file_hash=file_scan.file_hash, finish=last_timestamp))

        return mjournal, (first_timestamp, last_timestamp, mjournal, records)

    def _load_data(self, s, loader, data):
        first_timestamp, last_timestamp, mjournal, records = data
        for record in records:
            if HEART_RATE_ATTR in record.data and record.data[HEART_RATE_ATTR][0][0]:
                loader.add(HEART_RATE, BPM, None, ALL, mjournal, record.data[HEART_RATE_ATTR][0][0],
                           record.timestamp, StatisticJournalInteger,
                           description='''The instantaneous heart rate.''')
            if STEPS_ATTR in record.data:
                for (sport, steps) in zip(record.data[ACTIVITY_TYPE_ATTR][0], record.data[STEPS_ATTR][0]):
                    try:
                        loader.add(CUMULATIVE_STEPS, STEPS_UNITS, None,
                                   self.sport_to_activity_group[sport], mjournal, steps,
                                   record.timestamp, StatisticJournalInteger,
                                   description='''The number of steps in a day to this point in time.''')
                    except KeyError:
                        raise FatalException(f'There is no group configured for {sport} entries in the FIT file.')

    def _shutdown(self, s):
        super()._shutdown(s)
        if not self.worker:
            for activity_group_id in self._step_activity_group_ids(s):
                df = self._read_diff(s, activity_group_id)
                df = self._calculate_diff(df)
                self._write_diff(s, df, activity_group_id)

    def _step_activity_group_ids(self, s):
        return [row[0] for row in s.query(distinct(ActivityGroup.id)).
            join(StatisticName).
            filter(StatisticName.name == CUMULATIVE_STEPS,
                   StatisticName.owner == self.owner_out).all()]

    def _read_diff(self, s, activity_group_id):
        t = _tables()
        qs = select([t.sj.c.time.label("time"), t.sji.c.value.label("steps")]). \
            select_from(t.sj.join(t.sn).join(t.sji)). \
            where(and_(t.sn.c.name == STEPS, t.sn.c.activity_group_id == activity_group_id,
                       t.sn.c.owner == self.owner_out)).alias("steps")
        q = select([t.sj.c.time.label(TIME), t.sj.c.source_id.label(SOURCE), t.sji.c.value.label(CUMULATIVE_STEPS),
                    qs.c.steps.label(STEPS)]). \
            select_from(t.sj.join(t.sn).join(t.sji).outerjoin(qs, t.sj.c.time == qs.c.time)). \
            where(and_(t.sn.c.name == CUMULATIVE_STEPS, t.sn.c.activity_group_id == activity_group_id,
                       t.sn.c.owner == self.owner_out)). \
            order_by(t.sj.c.time)
        # log.debug(q)
        df = pd.read_sql_query(sql=q, con=s.connection(), index_col=TIME)
        return df

    def _calculate_diff(self, df):
        df[NEW_STEPS] = df[CUMULATIVE_STEPS].diff()
        df.loc[df[NEW_STEPS] < 0, NEW_STEPS] = df[CUMULATIVE_STEPS]
        df.loc[df[NEW_STEPS].isna(), NEW_STEPS] = df[CUMULATIVE_STEPS]
        return df

    def _write_diff(self, s, df, activity_group_id):
        activity_group = s.query(ActivityGroup).filter(ActivityGroup.id == activity_group_id).one()
        steps = StatisticName.add_if_missing(s, STEPS, StatisticJournalType.INTEGER, STEPS_UNITS, None,
                                             self.owner_out, activity_group, description=STEPS_DESCRIPTION)
        times = df.loc[(df[NEW_STEPS] != df[STEPS]) & ~df[STEPS].isna()].index.astype(np.int64) / 1e9
        if len(times):
            n = s.query(func.count(StatisticJournal.id)). \
                filter(StatisticJournal.time.in_(times),
                       StatisticJournal.statistic_name == steps).scalar()
            log.warning(f'Deleting {n} {STEPS}/{activity_group} entries')
            s.query(StatisticJournal.id). \
                filter(StatisticJournal.time.in_(times),
                       StatisticJournal.statistic_name == steps).delete(synchronize_session=False)
        loader = StatisticJournalLoader(s, owner=self.owner_out)
        for time, row in df.loc[(df[NEW_STEPS] != df[STEPS]) & ~df[NEW_STEPS].isna()].iterrows():
            loader.add(STEPS, STEPS_UNITS, None, activity_group, row[SOURCE], int(row[NEW_STEPS]),
                       time, StatisticJournalInteger, description=STEPS_DESCRIPTION)
        loader.load()




from logging import getLogger

from .calculate import MultiProcCalculator, ActivityJournalCalculatorMixin, DataFrameCalculatorMixin
from ...data.elevation import smooth_elevation
from ...data.frame import activity_statistics, present
from ...names import RAW_ELEVATION, ELEVATION, DISTANCE, M, GRADE, PC, ALTITUDE
from ...sql import StatisticJournalFloat

log = getLogger(__name__)


class ElevationCalculator(ActivityJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    def __init__(self, *args, smooth=3, **kargs):
        self.smooth = smooth
        super().__init__(*args, **kargs)

    def _read_dataframe(self, s, ajournal):
        try:
            df = activity_statistics(s, DISTANCE, RAW_ELEVATION, ELEVATION, ALTITUDE,
                                     activity_journal=ajournal, with_timespan=True)
            return df
        except Exception as e:
            log.warning(f'Failed to generate statistics for elevation: {e}')
            raise

    def _calculate_stats(self, s, ajournal, df):
        if not present(df, ELEVATION):
            if present(df, RAW_ELEVATION):
                df = smooth_elevation(df, smooth=self.smooth)
            elif present(df, ALTITUDE):
                log.warning(f'Using {ALTITUDE} as {ELEVATION}')
                df[ELEVATION] = df[ALTITUDE]
            return df
        else:
            return None

    def _copy_results(self, s, ajournal, loader, df):
        for time, row in df.iterrows():
            if ELEVATION in row:
                loader.add(ELEVATION, M, None, ajournal.activity_group, ajournal, row[ELEVATION], time,
                           StatisticJournalFloat,
                           description='An estimate of elevation (may come from various sources).')
            if GRADE in row:
                loader.add(GRADE, PC, None, ajournal.activity_group, ajournal, row[GRADE], time,
                           StatisticJournalFloat,
                           description='The gradient of the smoothed SRTM1 elevation.')




from itertools import groupby
from logging import getLogger

from math import sqrt, ceil
from pygeotile.point import Point

from .activity import ActivityReader
from .waypoint import filter_none
from ...lib.date import to_time, format_time
from ...lib.utils import sign
from ...rtree import MatchType
from ...rtree.spherical import LocalTangent, SQRTree, Global
from ...sql.database import Timestamp
from ...sql.tables.activity import ActivityGroup
from ...sql.tables.segment import Segment, SegmentJournal
from ...sql.utils import add

log = getLogger(__name__)

NAMES = {'Latitude': 'lat',
         'Longitude': 'lon',
         'Distance': 'distance'}


class CalcFailed(Exception): pass


class SegmentReader(ActivityReader):

    def __init__(self, *args, inner_bound=10, match_bound=25, **kargs):
        self.inner_bound = inner_bound
        self.match_bound = match_bound
        super().__init__(*args, **kargs)

    def _startup(self, s):
        SegmentJournal.clean(s)
        super()._startup(s)
        self.__segments = {}
        for agroup in s.query(ActivityGroup).all():
            self.__segments[agroup.id] = self._read_segments(s, agroup)

    def _shutdown(self, s):
        SegmentJournal.clean(s)
        super()._shutdown(s)

    def _load_data(self, s, loader, data):
        super()._load_data(s, loader, data)
        ajournal, activity_group, first_timestamp, file_scan, define, records = data
        self._find_segments(s, ajournal, filter_none(NAMES.values(), loader.as_waypoints(NAMES)))

    def _find_segments(self, s, ajournal, waypoints):
        matches = sorted(self._initial_matches(s, ajournal.activity_group_id, waypoints), key=lambda m: m[2].name)
        for segment, segment_matches in groupby(matches, key=lambda m: m[2]):
            log.debug(f'Considering {segment.name}')
            ordered = sorted(segment_matches, key=lambda m: m[0])
            coallesced = list(self._coallesce(ordered))
            starts, finishes, segment = self._split(coallesced)
            while starts:
                start = starts.pop(-1)  # work backwards through starts
                copy = list(finishes)
                while copy:
                    finish = copy.pop(0)  # work forwards through finishes
                    if finish[0] > start[0]:
                        if self._try_segment(s, start, finish, waypoints, segment, ajournal):
                            copy = None  # exit search for this start
                            # move to a start that won't overlap (working backwards)
                            while starts and starts[-1][0] > start[0] - 0.5 * (finish[0] - start[0]):
                                starts.pop(-1)
                            if starts:
                                log.debug('Possible second segment')

    def _try_segment(self, s, starts, finishes, waypoints, segment, ajournal):
        try:
            log.info('Trying segment %s-%s for %s' % (starts, finishes, segment.name))
            d = waypoints[self._mid(finishes)].distance - waypoints[self._mid(starts)].distance
            if abs(d - segment.distance) / segment.distance > 0.2:
                raise CalcFailed('Distance between start and finish (%.1fkm) doesn\'t match segment (%.1fkm)' %
                                 (d / 1000, segment.distance / 1000))
            start_time = self._end_point(starts, waypoints, segment.start, self.inner_bound, True)
            finish_time = self._end_point(finishes, waypoints, segment.finish, self.inner_bound, False)
            sjournal = add(s, SegmentJournal(segment_id=segment.id, activity_journal=ajournal,
                                             start=start_time, finish=finish_time))
            log.info('Added %s for %s - %s' %
                           (segment.name, format_time(start_time), format_time(finish_time)))
            s.commit()  # needed to get id on sjournal (and let other workers in)
            # since source is sjournal we cannot use on_success
            Timestamp.set(s, self, constraint=segment, source=sjournal)
            return True
        except CalcFailed as e:
            log.warning(str(e))
            return False

    def _mid(self, indices):
        n = len(indices)
        return indices[n // 2]

    def _end_point(self, indices, waypoints, p, inner, hi_to_lo):
        '''
        Find the time of the point that makes the segment shortest while also being at a local
        minimum in distance from the start/finish point that is within inner m.
        '''
        metric = LocalTangent(p)
        lo, hi = self._limits(metric, waypoints, indices, p)
        log.info('Finding closest point to %s within %d: %f, %d: %f' %
                       (p, lo, self._dw(metric, p, waypoints, lo),
                        hi, self._dw(metric, p, waypoints, hi)))
        if hi_to_lo:
            start, finish = hi, lo
        else:
            start, finish = lo, hi
        minimum = self._minimum(metric, waypoints, start, finish, p, inner)
        # interpolate to fractional index
        i0, i1 = int(minimum), int(minimum) + 1
        k = (minimum - i0) / (i1 - i0)
        t0, t1 = waypoints[i0].time.timestamp(), waypoints[i1].time.timestamp()
        return to_time(t0 * (1 - k) + t1 * k)

    def _plot(self, p):
        '''
        Plot points for debugging
        '''
        lon, lat = p
        print('PLOT: %s' % (Point.from_latitude_longitude(lat, lon).meters, ))

    def _plotw(self, metric, waypoints, i, d):
        '''
        Plot points for debugging
        '''
        i1, i2 = int(i), int(i) + 1
        x1, y1 = metric.normalize((waypoints[i1].lon, waypoints[i1].lat))
        x2, y2 = metric.normalize((waypoints[i2].lon, waypoints[i2].lat))
        k = i - i1
        xk = x1 * (1 - k) + x2 * k
        yk = y1 * (1 - k) + y2 * k
        lon, lat = metric.denormalize((xk, yk))
        print('PLOT: %s %.1f' % (Point.from_latitude_longitude(lat, lon).meters, d))

    def _minimum(self, metric, waypoints, start, finish, p, inner):
        '''
        Find the (fractional) index of the point with minimum distance to the endpoint within the two limits.
        '''
        # self._plot(p)
        i = start
        while True:
            i, min_d = self._next_local_minimum(metric, p, waypoints, i, start, finish)
            log.info('Local minimum %.1f: %f (< %.1f?)' % (i, min_d, inner))
            # self._plotw(metric, waypoints, i, min_d)
            if min_d < inner:
                return i
            # move to next segment
            if start < finish:
                i = int(i + 1)
            else:
                i = ceil(i - 1)

    def _next_local_minimum(self, metric, p, waypoints, i, start, finish):
        '''
        Find the next point with a minimum distance to p, along the waypoints (interpolated),
        starting from i and between start and finish (moving towards finish).
        '''
        # this from basic algebra (derivation too long for a comment but pretty simple)
        # p0 is the point we want to minimize distance to
        # pi and pj are the end points of a nearby line segment
        # k is the fractional distance between pi and pj for the point nearest to p0
        # p1 is the point at k if it is between pi and pj
        # things are complicated by the need to return pj if it's the nearest point (discontinuity)
        p0 = metric.normalize(p)
        pi = metric.normalize((waypoints[i].lon, waypoints[i].lat))
        prev_discontinuity = None
        while start <= i < finish or start >= i > finish:
            j = i + sign(finish - start)
            pj = metric.normalize((waypoints[j].lon, waypoints[j].lat))
            dxji, dyji = pj[0] - pi[0], pj[1] - pi[1]
            dx0i, dy0i = p0[0] - pi[0], p0[1] - pi[1]
            k = (dxji * dx0i + dyji * dy0i) / (dxji**2 + dyji**2)
            if 0 < k < 1:  # we have a solution within the segment
                p1 = (pi[0] + k * (pj[0] - pi[0]), pi[1] + k * (pj[1] - pi[1]))
                return i + k * sign(finish - start), sqrt((p0[0] - p1[0])**2 + (p0[1] - p1[1])**2)
            # are we moving towards the target?
            if k > 1:
                # if so, save this endpoint as a possible minimum
                prev_discontinuity = (j, self._dw(metric, p, waypoints, j))
            else:
                # if not, then if we were before it was a minimum
                if prev_discontinuity:
                    return prev_discontinuity
            i, pi = j, pj
        raise CalcFailed('No minimum found')

    def _limits(self, metric, waypoints, indices, p):
        '''
        Given a pair of indices, move them apart so that the distance from the endpoint has increased
        in both directions (ie they bracket the endpoint in some sense).
        '''
        lo, hi = indices
        if lo == hi:
            lo, hi = lo-1, hi+1
        dl, dh = self._dw(metric, p, waypoints, lo), self._dw(metric, p, waypoints, hi)
        if dl < dh:
            lo, dl = self._inc_limit(metric, waypoints, p, dl, dh, lo, -1)
            hi, dh = self._inc_limit(metric, waypoints, p, dh, dl, hi, 1)
        else:
            hi, dh = self._inc_limit(metric, waypoints, p, dh, dl, hi, 1)
            lo, dl = self._inc_limit(metric, waypoints, p, dl, dh, lo, -1)
        log.info('Expanded %s to %s' % (indices, (lo, hi)))
        return lo, hi

    def _inc_limit(self, metric, waypoints, p, da, db, a, inc):
        '''
        Move one index.
        '''
        while da < db and ((inc > 0 and a < len(waypoints) - 1) or (inc < 0 and a > 0)):
            a += inc
            da = self._dw(metric, p, waypoints, a)
        return a, da

    def _dw(self, metric, p, waypoints, i):
        '''
        Distance between an waypoint and the endpoint.
        '''
        return self._d(metric, p, (waypoints[i].lon, waypoints[i].lat))

    def _d(self, metric, p1, p2):
        '''
        The distance between two points using the given metric.
        '''
        x1, y1 = metric.normalize(p1)
        x2, y2 = metric.normalize(p2)
        return sqrt((x1 - x2)**2 + (y1 - y2)**2)

    def _split(self, coallesced):
        '''
        Split waypoints for a single segment into separate start and finish points.
        '''
        starts, finishes, segment = [], [], None
        for i, start, segment in coallesced:
            if start:
                starts.append(i)
            else:
                finishes.append(i)
        return starts, finishes, segment

    def _coallesce(self, ordered_sm):
        '''
        Replace contiguous ranges of waypoints with a pair (min, max)
        '''
        prev, first = None, None
        for match in ordered_sm:
            if match != prev:  # skip duplicates
                if prev and (prev[1:2] != match[1:2] or prev[0]+1 != match[0]):
                    yield (first, prev[0]), prev[1], prev[2]
                    prev, first = None, None
                if first is None:
                    first = match[0]
            prev = match
        yield (first, prev[0]), prev[1], prev[2]

    def _initial_matches(self, s, agroup_id, waypoints):
        '''
        Check each waypoint against the r-tree and return all matches.
        '''
        found = set()
        for i, waypoint in enumerate(waypoints):
            for start, segment in self.__segments[agroup_id][[(waypoint.lon, waypoint.lat)]]:
                if segment not in found:
                    log.info(f'Candidate segment "{segment.name}" at ({segment.start_lon}, {segment.start_lat}) - '
                             f'({segment.finish_lon}, {segment.finish_lat})')
                    found.add(segment)
                log.debug(f'Match at {(waypoint.lon, waypoint.lat)} '
                          f'for {segment.name} {"start" if start else "finish"}')
                yield i, start, segment

    def _read_segments(self, s, agroup):
        '''
        Read segment endpoints into a global R-tree so we can detect when waypoints pass nearby.
        '''
        segments = Global(tree=lambda: SQRTree(default_border=self.match_bound, default_match=MatchType.OVERLAP))
        for segment in s.query(Segment).all():
            segments[[segment.start]] = (True, segment)
            segments[[segment.finish]] = (False, segment)
        if not segments:
            log.warning('No segments defined in database for %s' % agroup)
        return segments
